import java.awt.*;
import java.io.*;

public class Selector
{
  public static int stateMask = 31;
  public static int selected = 32;

  public enum State {None, TopLeftMarked, GroupSelected};

  private HexagonalCell hexagonalCell;
  private HexGrid hexGrid;
  private int cells[][];
  private int worldColumns;
  private int worldRows;
  private int selection[][];
  private State state;
  private int topLeftColumn;
  private int topLeftRow;
  private int columns;
  private int rows;
  private boolean cut;
  private boolean evenOrientation;
  private int even[][];
  private int evenColumns;
  private int odd[][];
  private int oddColumns;

  public Selector(HexagonalCell hexagonalCell, HexGrid hexGrid, int cells[][], int noColumns, int noRows)
  {
    this.hexagonalCell = hexagonalCell;
    this.hexGrid = hexGrid;
    this.cells = cells;
    worldColumns = noColumns;
    worldRows = noRows;
    state = State.None;
    cut = true;
    evenOrientation = true;
  }

  public void setCells(int[][] cells)
  {
    this.cells = cells;
    if (state == State.GroupSelected)
      markSelectedCells();
  }

  public State getState()
  {
    return state;
  }

  public void setCut(boolean cut)
  {
    this.cut = cut;
  }

  public boolean getCut()
  {
    return cut;
  }

  public int getCell(int i, int j)
  {
    return selection[(worldColumns + i - topLeftColumn) % worldColumns][(worldRows + j - topLeftRow) % worldRows];
  }

  public void setTopLeft(int c, int r)
  {
    Point p = hexGrid.selectCell(c, r);
    if (p == null)
      return;
    state = State.TopLeftMarked;
    topLeftColumn = p.x;
    topLeftRow = p.y;
    makeNewSelection(p.x, p.y);
    selectCells(false);
    hexGrid.setDrawSelector(true);
    hexGrid.repaint();
  }

  private void makeNewSelection(int column, int row)
  {
    int t;
    int c = topLeftColumn;
    int r = topLeftRow;
    if (c > column)
    {
      t = c;
      c = column;
      column = t;
    }
    if (r > row)
    {
      t = r;
      r = row;
      row = t;
    }
    topLeftColumn = c;
    topLeftRow = r;
    columns = column - topLeftColumn + 1;
    rows = row - topLeftRow + 1;
    selection = new int[columns][rows];
  }

  private void selectCells(boolean cut)
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
    {
      int c = hexGrid.offsetColumn(i);
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
      {
        int r = hexGrid.offsetRow(j);
        selection[i - topLeftColumn][j - topLeftRow] = cells[c][r] & stateMask;
        if (cut)
          cells[c][r] = selected;
        else
          cells[c][r] |= selected;
      }
    }
  }

  private void setOrientation()
  {
    evenOrientation = true;
    if ((topLeftRow & 1) == 1)
      evenOrientation = false;
  }

  public void save(int c, int r)
  {
    Point p = hexGrid.selectCell(c, r);
    if (p == null)
      return;
    hexGrid.setDrawSelector(false);
    state = State.GroupSelected;
    makeNewSelection(p.x, p.y);
    selectCells(cut);
    setOrientation();
    makeEvenAndOddCopies();
    hexGrid.setDrawSelector(true);
    hexGrid.repaint();
  }

  private void clearSelectedBits()
  {
    for (int i = 0; i < worldColumns; i++)
      for (int j = 0; j < worldRows; j++)
        cells[i][j] &= stateMask;
  }

  public void restoreCells()
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
    {
      int c = hexGrid.offsetColumn(i);
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
      {
        int r = hexGrid.offsetRow(j);
        int v = selection[i - topLeftColumn][j - topLeftRow];
        if (v != 0)
          cells[c][r] = v;
      }
    }
  }

  public void remove()
  {
    hexGrid.setDrawSelector(false);
    state = State.None;
    clearSelectedBits();
    selection = null;
    even = null;
    odd = null;
    hexGrid.repaint();
  }

  public void clearOutside()
  {
    for (int c = 0; c < worldColumns; c++)
      for (int r = 0; r < worldRows; r++)
        cells[c][r] &= selected;
    hexGrid.repaint();
  }

  private void markSelectedCells()
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
        cells[hexGrid.offsetColumn(i)][hexGrid.offsetRow(j)] |= selected;
  }

  private void makeEvenAndOddCopies()
  {
    if (evenOrientation)
    {
      odd = new int[columns + 1][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          if ((r % 2) == 0)
            odd[c][r] = selection[c][r];
          else
            odd[c + 1][r] = selection[c][r];
      oddColumns = columns + 1;
      even = selection;
      evenColumns = columns;
    }
    else
    {
      even = new int[columns + 1][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          if ((r % 2) == 0)
            even[c + 1][r] = selection[c][r];
          else
            even[c][r] = selection[c][r];
      evenColumns = columns + 1;
      odd = selection;
      oddColumns = columns;
    }
  }

  private void selectEvenOrOdd()
  {
    if (((topLeftRow % 2) == 0) ^ evenOrientation)
    {
      if (evenOrientation)
      {
        selection = odd;
        columns = oddColumns;
      }
      else
      {
        selection = even;
        columns = evenColumns;
      }
      evenOrientation = ! evenOrientation;
    }
  }

  public boolean move(int c, int r)
  {
    Point p = hexGrid.selectCell(c, r);
    if (p == null)
      return false;
    hexGrid.setDrawSelector(false);
    topLeftColumn = p.x;
    topLeftRow = p.y;
    selectEvenOrOdd();
    clearSelectedBits();
    markSelectedCells();
    hexGrid.setDrawSelector(true);
    hexGrid.repaint();
    return true;
  }

  private void trimSelection()
  {
    int leftColumn = columns;
    int topRow = rows;
    int rightColumn = -1;
    int bottomRow = -1;
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        if (selection[c][r] != 0)
        {
          if (c < leftColumn)
            leftColumn = c;
          if (c > rightColumn)
            rightColumn = c;
          if (r < topRow)
            topRow = r;
          if (r > bottomRow)
            bottomRow = r;
        }
    int newColumns = rightColumn - leftColumn + 1;
    int newRows = bottomRow - topRow + 1;
    int newSelection[][] = new int[newColumns][newRows];
    for (int c = 0; c < newColumns; c++)
      for (int r = 0; r < newRows; r++)
        newSelection[c][r] = selection[c + leftColumn][r + topRow];
    selection = newSelection;
    columns = newColumns;
    rows = newRows;
    incTopLeftColumn(leftColumn);
    incTopLeftRow(topRow);
    setOrientation();
    makeEvenAndOddCopies();
    clearSelectedBits();
    markSelectedCells();
  }

  public void flipX()
  {
    if (state != State.GroupSelected)
      return;
    hexGrid.setDrawSelector(false);
    int t[][];
    if ((rows % 2) == 1)
    {
      t = new int[columns][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[c][rows - r - 1] = selection[c][r];
    }
    else
    {
      t = new int[columns][rows + 1];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[c][rows - r] = selection[c][r];
      rows++;
    }
    selection = t;
    trimSelection();
    hexGrid.setDrawSelector(true);
    center();
  }

  public void flipY()
  {
    if (state != State.GroupSelected)
      return;
    hexGrid.setDrawSelector(false);
    int t[][] = new int[columns + 1][rows];
    for (int r = 0; r < rows; r++)
      if (!(((topLeftRow % 2) == 0) ^ ((r % 2) == 0)))
        for (int c = 0; c < columns; c++)
          t[columns - c][r] = selection[c][r];
      else
        for (int c = 0; c < columns; c++)
          t[columns - c - 1][r] = selection[c][r];
    columns++;
    selection = t;
    trimSelection();
    hexGrid.setDrawSelector(true);
    center();
  }

  public void rotateClockwise()
  {
    if (state != State.GroupSelected)
      return;
    hexGrid.setDrawSelector(false);
    int newColumns, newRows;
    int sc, sr, nc, nr, ca, ra;
    sr = 0;
    if (evenOrientation)
    {
      sc = (rows / 4) * 3 + rows % 4 - 1;
      newColumns = (columns - 1) / 2  + sc + 1;
      newRows = columns + rows / 2;
      ca = 0;
      ra = 2;
    }
    else
    {
      sc = ((rows + 1) / 4) * 3 + (rows + 1) % 4 - 2;
      newColumns = columns / 2 + sc + 1;
      newRows = columns + (rows - 1) / 2;
      ca = 1;
      ra = 3;
    }
    int t[][] = new int[newColumns][newRows];
    nc = sc;
    nr = sr;
    for (int r = 0; r < rows; r++)
    {
      for (int c = 0; c < columns; c++)
      {
        t[nc][nr] = selection[c][r];
        nc += (c + ca) % 2;
        nr++;
      }
      if (((r + ra) % 4) != 0)
        sc--;
      sr += (r + ra + 1) % 2;
      if (!((sr % 2) == 0) ^ evenOrientation)
        ca = 0;
      else
        ca = 1;
      nc = sc;
      nr = sr;
    }
    selection = t;
    columns = newColumns;
    rows = newRows;
    trimSelection();
    hexGrid.setDrawSelector(true);
    center();
  }

  public void rotateAntiClockwise()
  {
    for (int i = 0; i < 5; i++)
      rotateClockwise();
    center();
  }

  public void selectAll()
  {
    if (state != State.None)
    {
      restoreCells();
      remove();
    }
    int c1 = worldColumns;
    int r1 = worldRows;
    int cn = 0;
    int rn = 0;
    boolean clear = true;
    for (int c = 0; c < worldColumns; c++)
      for (int r = 0; r < worldRows; r++)
        if (cells[c][r] > 0)
        {
          clear = false;
          if (c < c1)
            c1 = c;
          if (c > cn)
            cn = c;
          if (r < r1)
            r1 = r;
          if (r > rn)
            rn = r;
        }
    if (! clear)
    {
      topLeftColumn = hexGrid.worldColumnToGridColumn(c1);
      topLeftRow = hexGrid.worldRowToGridRow(r1);
      columns = hexGrid.worldColumnToGridColumn(cn) - topLeftColumn + 1;
      if (columns <= 0)
        columns += worldColumns;
      rows = hexGrid.worldRowToGridRow(rn) - topLeftRow + 1;
      if (rows <= 0)
        rows += worldRows;
      selection = new int[columns][rows];
      selectCells(cut);
      state = State.GroupSelected;
      setOrientation();
      makeEvenAndOddCopies();
      hexGrid.setDrawSelector(true);
      hexGrid.repaint();
    }
  }

  public void center()
  {
    if ((state != State.GroupSelected) || (columns > hexGrid.getGridColumns()) || (rows > hexGrid.getGridRows()))
      return;
    hexGrid.setDrawSelector(false);
    topLeftColumn = (hexGrid.getGridColumns() - columns) / 2;
    topLeftRow = (hexGrid.getGridRows() - rows) / 2;
    selectEvenOrOdd();
    clearSelectedBits();
    markSelectedCells();
    hexGrid.setDrawSelector(true);
    hexGrid.repaint();
  }

  public void incTopLeftColumn(int inc)
  {
    topLeftColumn = (topLeftColumn + inc) % worldColumns;
  }

  public void decTopLeftColumn(int inc)
  {
    topLeftColumn = (worldColumns + topLeftColumn - inc) % worldColumns;
  }

  public void incTopLeftRow(int inc)
  {
    topLeftRow = (topLeftRow + inc) % worldRows;
  }

  public void decTopLeftRow(int inc)
  {
    topLeftRow = (worldRows + topLeftRow - inc) % worldRows;
  }

  public void readNewSelection(BufferedReader br)
  {
    if (state != State.None)
    {
      restoreCells();
      remove();
    }
    try
    {
      rows = new Integer(br.readLine().substring(8).trim());
      columns = new Integer(br.readLine().substring(11).trim());
      if (br.readLine().substring(15).equals("Even"))
        evenOrientation = true;
      else
        evenOrientation = false;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    topLeftColumn = 0;
    topLeftRow = 0;
    selection = new int[columns][rows];
    readSelection(br);
    state = State.GroupSelected;
    markSelectedCells();
    makeEvenAndOddCopies();
    center();
  }

  private void readSelection(BufferedReader br)
  {
    int r, c;
    int n;
    char ch;
    String s = "";
    try
    {
      String line = br.readLine();
      while (line != null)
        if (line.startsWith("#L "))
        {
          s = s + line.substring(3);
          line = br.readLine();
        }
        else
          return;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    r = 0;
    c = 0;
    while (s.length() > 0)
    {
      n = 0;
      ch = s.charAt(0);
      while ((ch >= '0') && (ch <= '9'))
      {
        n = n * 10 + (ch - '0');
        s = s.substring(1);
        ch = s.charAt(0);
      }
      if (n == 0)
        n = 1;
      if (ch == '$')
      {
        r += n;
        c = 0;
      }
      else
        if (ch == '.')
          c += n;
        else
        {
          for (int i = 0;  i < n; i++)
          {
            selection[c][r] = (byte)(ch - '@');
            c++;
          }
        }
      s = s.substring(1);
    }
  }

  public void writeHeader(PrintWriter pw)
  {
    pw.println("#Rows = " + rows);
    pw.println("#Columns = " + columns);
    if (evenOrientation)
      pw.println("#Orientation = Even");
    else
      pw.println("#Orientation = Odd");
  }

  public void writeSelection(PrintWriter pw)
  {
    int r, c;
    int ce = columns - 1;
    char state;
    char lastState = '0';
    int n = 0;
    boolean blankLine;
    String s = "";
    for (r = 0; r < rows; r++)
    {
      blankLine = true;
      for (c = 0; c < columns; c++)
      {
        if (selection[c][r] > 0)
        {
          blankLine = false;
          break;
        }
      }
      if (blankLine)
      {
        if (lastState == '$')
          n++;
        else
        {
          lastState = '$';
          n = 1;
        }
        continue;
      }
      else
        for (c = columns - 1; c >= 0; c--)
        {
          if (selection[c][r] > 0)
          {
            ce = c;
            break;
          }
        }
      for (c = 0; c <= ce;  c++)
      {
        state = (char)(selection[c][r] + 64);
        if (state == '@')
          state = '.';
        if (n == 0)
        {
          lastState = state;
          n = 1;
        }
        else if (state == lastState)
          n++;
        else
        {
          if (n > 1)
            s = s + n;
          s = s + lastState;
          if (s.length() >= 67)
          {
            pw.println("#L " + s);
            s = "";
          }
          lastState = state;
          n = 1;
        }
      }
      if (lastState == '$')
        n++;
      else
      {
        if (n > 1)
          s = s + n;
        s = s + lastState;
        if (s.length() >= 67)
        {
          pw.println("#L " + s);
          s = "";
        }
        lastState = '$';
        n = 1;
      }
    }
    if (s != "")
      pw.println("#L " + s);
  }
  
  public void random()
  {
    if (state != State.GroupSelected)
      return;
    hexagonalCell.random(columns, rows, selection);
    makeEvenAndOddCopies();
  }
}