import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

class CountsRule extends Rule
{
  private static final int neighborRows = 5;
  private static final int neighborColumns = 5;

  private int noStates;
  private int percentChange = 30;
  private boolean clearNewRule = true;
  private int neighborWeights[][];
  private int ruleTable[][];

  public CountsRule(JFrame mainFrame, SquareCell squareCell)
  {
    super(mainFrame, squareCell);
    setDefault();
  }

  private void makeArrays()
  {
    neighborWeights = new int[neighborRows][neighborColumns];
    ruleTable = new int[noStates - 1][25];
  }

  private void setWeights()
  {
    for (int r = 0; r < neighborRows; r++)
      for (int c = 0; c < neighborColumns; c++)
        if ((r > 0) && (r < 4) && (c > 0) && (c < 4))
          neighborWeights[r][c] = 1;
        else
          neighborWeights[r][c] = 0;
    neighborWeights[2][2] = 0;
  }

  public void setNewRule()
  {
    random(ruleTable);
  }

  public void random(int ruleTable[][])
  {
    if (clearNewRule)
    {
      for (int r = 0; r < noStates - 1; r++)
        for (int c = 0; c < 25; c++)
          ruleTable[r][c] = 0;
    }
    Random generator = new Random();
    for (int i = 1; i < (noStates - 1) * 25 * percentChange / 100; i++)
    {
      int r = generator.nextInt(noStates - 1);
      int c = generator.nextInt(25);
      ruleTable[r][c] = generator.nextInt(noStates);
    }
  }

  public void setOptions()
  {
    int currentNoStates = noStates;
    CountsOptionsDialog countsOptionsDialog = new CountsOptionsDialog(mainFrame);
    countsOptionsDialog.setData(this);
    centerDialog(countsOptionsDialog);
    countsOptionsDialog.setVisible(true);
    if (countsOptionsDialog.hasChanged())
    {
      countsOptionsDialog.getData(this);
      if (noStates != currentNoStates)
        ruleTable = new int[noStates][25];
    }
  }

  public void setDefault()
  {
    noStates = 16;
    makeArrays();
    setWeights();
  }

  public void changeRule()
  {
    CountsDialog countsDialog = new CountsDialog(mainFrame, this, squareCell);
    countsDialog.setData(this);
    centerDialog(countsDialog);
    countsDialog.setVisible(true);
    if (countsDialog.hasChanged())
      countsDialog.getData(this);
  }

  private void centerDialog(JDialog dialog)
  {
    Point p = mainFrame.getLocation();
    int x = (int)p.getX() + (mainFrame.getWidth() - dialog.getWidth()) / 2;
    int y = (int)p.getY() + (mainFrame.getHeight() - dialog.getHeight()) / 2;
    dialog.setLocation(x, y);
  }

  public void writeRule(PrintWriter pw)
  {
    String s;
    pw.println("#Rule = Counts");
    pw.println("#States = " + noStates);
    for (int i = 0; i < 5; i++)
    {
      s = "#NW ";
      for (int j = 0; j < 5; j++)
        s = s + neighborWeights[i][j] + ",";
      pw.println(s.substring(0, s.lastIndexOf(",")));
    }
    for (int i = 0; i < noStates - 1; i++)
    {
      s = "#RT ";
      for (int j = 0; j < 25; j++)
        s = s + ruleTable[i][j] + ",";
      pw.println(s.substring(0, s.lastIndexOf(",")));
    }
  }

  public void readRule(BufferedReader br)
  {
    try
    {
      noStates = new Integer(br.readLine().substring(10).trim());
      ruleTable = new int[noStates - 1][25];
      for (int i = 0; i < 5; i++)
      {
        String values[] = br.readLine().substring(3).split(",");
        for (int j = 0; j < 5; j++)
          neighborWeights[i][j] = new Integer(values[j].trim());
      }
      for (int i = 0; i < noStates - 1; i++)
      {
        String values[] = br.readLine().substring(3).split(",");
        for (int j = 0; j < 25; j++)
          ruleTable[i][j] = new Integer(values[j].trim());
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  public int step(int r, int c)
  {
    int count =
      squareCell.getNeighbor(r - 2, c - 2) * neighborWeights[0][0] +
      squareCell.getNeighbor(r - 2, c - 1) * neighborWeights[0][1] +
      squareCell.getNeighbor(r - 2, c    ) * neighborWeights[0][2] +
      squareCell.getNeighbor(r - 2, c + 1) * neighborWeights[0][3] +
      squareCell.getNeighbor(r - 2, c + 2) * neighborWeights[0][4] +

      squareCell.getNeighbor(r - 1, c - 2) * neighborWeights[1][0] +
      squareCell.getNeighbor(r - 1, c - 1) * neighborWeights[1][1] +
      squareCell.getNeighbor(r - 1, c    ) * neighborWeights[1][2] +
      squareCell.getNeighbor(r - 1, c + 1) * neighborWeights[1][3] +
      squareCell.getNeighbor(r - 1, c + 2) * neighborWeights[1][4] +

      squareCell.getNeighbor(r    , c - 2) * neighborWeights[2][0] +
      squareCell.getNeighbor(r    , c - 1) * neighborWeights[2][1] +
      squareCell.getNeighbor(r    , c    ) * neighborWeights[2][2] +
      squareCell.getNeighbor(r    , c + 1) * neighborWeights[2][3] +
      squareCell.getNeighbor(r    , c + 2) * neighborWeights[2][4] +

      squareCell.getNeighbor(r + 1, c - 2) * neighborWeights[3][0] +
      squareCell.getNeighbor(r + 1, c - 1) * neighborWeights[3][1] +
      squareCell.getNeighbor(r + 1, c    ) * neighborWeights[3][2] +
      squareCell.getNeighbor(r + 1, c + 1) * neighborWeights[3][3] +
      squareCell.getNeighbor(r + 1, c + 2) * neighborWeights[3][4] +

      squareCell.getNeighbor(r + 2, c - 2) * neighborWeights[4][0] +
      squareCell.getNeighbor(r + 2, c - 1) * neighborWeights[4][1] +
      squareCell.getNeighbor(r + 2, c    ) * neighborWeights[4][2] +
      squareCell.getNeighbor(r + 2, c + 1) * neighborWeights[4][3] +
      squareCell.getNeighbor(r + 2, c + 2) * neighborWeights[4][4];
    if (count == 0)
      return 0;
    return ruleTable[(count - 1) / 25][(count - 1) % 25];
  }

  public int getNoStates()
  {
    return noStates;
  }

  public int getPercentChange()
  {
    return percentChange;
  }

  public boolean getClearNewRule()
  {
    return clearNewRule;
  }

  public int getNeighborWeight(int row, int column)
  {
    return neighborWeights[row][column];
  }

  public int getRuleTableValue(int row, int column)
  {
    return ruleTable[row][column];
  }

  public void setNoStates(int value)
  {
    noStates = value;
  }

  public void setPercentChange(int percentChange)
  {
    this.percentChange = percentChange;
  }

  public void setClearNewRule(boolean clearNewRule)
  {
    this.clearNewRule = clearNewRule;
  }

  public void setNeighborWeight(int row, int column, int value)
  {
    neighborWeights[row][column] = value;
  }

  public void setRuleTableValue(int row, int column, int value)
  {
    ruleTable[row][column] = value;
  }
}
