import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

class InteractingStatesRule extends Rule
{
  private static final int noStates = 9;
  private static final int rows = 8;
  private static final int columns = 8;

  private boolean birthEnabled = true;
  private boolean surviveEnabled = false;
  private int birthPercentChange = 25;
  private int survivePercentChange = 25;
  private boolean birthClear = true;
  private boolean surviveClear = false;

  private int birth[][];
  private int survive[][];
  private int birthRandomLimit;
  private int surviveRandomLimit;
  private String birthOperatorSelector;
  private String surviveOperatorSelector;

  public InteractingStatesRule(JFrame mainFrame, SquareCell squareCell)
  {
    super(mainFrame, squareCell);
    setDefault();
  }

  public void setNewRule()
  {
    if (birthEnabled)
    {
      if (birthClear)
      {
        for (int r = 0; r < rows; r++)
          for (int c = 0; c < columns; c++)
            birth[r][c] = 0;
      }
      Random generator = new Random();
      for (int i = 0; i < rows * columns * birthPercentChange / 100; i++)
        birth[generator.nextInt(rows)][generator.nextInt(columns)] = generator.nextInt(birthRandomLimit + 1);
    }
    if (surviveEnabled)
    {
      if (surviveClear)
      {
        for (int r = 0; r < rows; r++)
          for (int c = 0; c < columns; c++)
            survive[r][c] = 0;
      }
      Random generator = new Random();
      for (int i = 0; i < rows * columns * survivePercentChange / 100; i++)
        survive[generator.nextInt(rows)][generator.nextInt(columns)] = generator.nextInt(surviveRandomLimit + 1);
    }
  }

  public void setOptions()
  {
    InteractingStatesOptionsDialog interactingStatesOptionsDialog = new InteractingStatesOptionsDialog(mainFrame);
    interactingStatesOptionsDialog.setData(this);
    centerDialog(interactingStatesOptionsDialog);
    interactingStatesOptionsDialog.setVisible(true);
    if (interactingStatesOptionsDialog.hasChanged())
      interactingStatesOptionsDialog.getData(this);
  }

  public void setDefault()
  {
    int[][] birth =
    {
      {0, 0, 0, 0, 2, 0, 3, 0},
      {0, 0, 0, 0, 0, 0, 0, 0},
      {0, 2, 0, 3, 0, 0, 0, 0},
      {0, 0, 0, 2, 0, 0, 0, 0},
      {0, 0, 2, 0, 0, 0, 1, 0},
      {0, 0, 1, 0, 1, 0, 0, 0},
      {0, 0, 0, 2, 0, 0, 0, 3},
      {0, 2, 0, 0, 0, 3, 0, 0}
    };
    this.birth = birth;
    int[][] survive =
    {
      {0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 1, 0, 0, 2, 0},
      {1, 0, 0, 0, 0, 0, 0, 0},
      {2, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 2, 0, 0, 0, 0, 0},
      {2, 0, 1, 0, 0, 0, 0, 0},
      {0, 0, 0, 2, 2, 0, 0, 0},
      {0, 0, 2, 0, 0, 0, 0, 1}
    };
    this.survive = survive;
    birthRandomLimit = 3;
    surviveRandomLimit = 2;
    birthOperatorSelector = "OR >=";
    surviveOperatorSelector = "OR =";
  }

  public void changeRule()
  {
    InteractingStatesDialog interactingStatesDialog = new InteractingStatesDialog(mainFrame);
    interactingStatesDialog.setData(this);
    centerDialog(interactingStatesDialog);
    interactingStatesDialog.setVisible(true);
    if (interactingStatesDialog.hasChanged())
      interactingStatesDialog.getData(this);
  }

  private void centerDialog(JDialog dialog)
  {
    Point p = mainFrame.getLocation();
    int x = (int)p.getX() + (mainFrame.getWidth() - dialog.getWidth()) / 2;
    int y = (int)p.getY() + (mainFrame.getHeight() - dialog.getHeight()) / 2;
    dialog.setLocation(x, y);
  }

  private String makeDataString(int[][] optionGrid, int randomLimit, String operatorSelector)
  {
    String s = "";
    for (int r = 0; r < rows; r++)
    {
      for (int c = 0; c <columns; c++)
        s += optionGrid[r][c];
      s += ",";
    }
    s += randomLimit + "," + operatorSelector;
    return s;
  }

  private void setRule(String[] s, int[][] rule)
  {
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        rule[r][c] = (int)(s[r].charAt(c) - '0');
  }

  public void writeRule(PrintWriter pw)
  {
    pw.println("#Rule = Interacting States");
    String birthData = makeDataString(birth, birthRandomLimit, birthOperatorSelector);
    String surviveData = makeDataString(survive, surviveRandomLimit, surviveOperatorSelector);
    pw.println("#Birth   = " + birthData);
    pw.println("#Survive = " + surviveData);
  }

  public void readRule(BufferedReader br)
  {
    try
    {
      String sb[] = br.readLine().substring(11).split(",");
      setRule(sb, birth);
      birthRandomLimit = (int)(sb[8].charAt(0) - '0');
      birthOperatorSelector = sb[9];
      String ss[] = br.readLine().substring(11).split(",");
      setRule(ss, survive);
      surviveRandomLimit = (int)(ss[8].charAt(0) - '0');
      surviveOperatorSelector = ss[9];
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  public int step(int row, int column)
  {
    int neighbors[] =
    {
      squareCell.getNeighbor(row - 1, column - 1),
      squareCell.getNeighbor(row    , column - 1),
      squareCell.getNeighbor(row + 1, column - 1),
      squareCell.getNeighbor(row - 1, column    ),
      squareCell.getNeighbor(row + 1, column    ),
      squareCell.getNeighbor(row - 1, column + 1),
      squareCell.getNeighbor(row    , column + 1),
      squareCell.getNeighbor(row + 1, column + 1)
    };
    int cell = squareCell.getNeighbor(row, column);
    int counts[] = new int[columns];
    for (int i = 0; i < columns; i++)
      if (neighbors[i] > 0)
        counts[neighbors[i] - 1]++;
    if (cell == 0)
    {
      if (birthOperatorSelector.equals("OR >="))
      {
        for (int r = 0; r < rows; r++)
          for (int c = 0; c < columns; c++)
            if ((birth[r][c] > 0) && (counts[c] >= birth[r][c]))
              return r + 1;
      }
      else if (birthOperatorSelector.equals("OR ="))
      {
        for (int r = 0; r < rows; r++)
          for (int c = 0; c < columns; c++)
            if ((birth[r][c] > 0) && (counts[c] == birth[r][c]))
              return r + 1;
      }
      else if (birthOperatorSelector.equals("AND >="))
      {
        for (int r = 0; r < rows; r++)
        {
          int n = 0;
          for (int c = 0; c < columns; c++)
            if (birth[r][c] > 0)
            {
              n = r + 1;
              break;
            }
          if (n == 0)
            break;
          for (int c = 0; c < columns; c++)
            if ((birth[r][c] > 0) && (counts[c] < birth[r][c]))
            {
              n = 0;
              break;
            }
          if (n > 0)
            return n;
        }
      }
      else if (birthOperatorSelector.equals("AND ="))
      {
        for (int r = 0; r < rows; r++)
        {
          int n = 0;
          for (int c = 0; c < columns; c++)
            if (birth[r][c] > 0)
            {
              n = r + 1;
              break;
            }
          if (n == 0)
            break;
          for (int c = 0; c < columns; c++)
            if ((birth[r][c] > 0) && (counts[c] != birth[r][c]))
            {
              n = 0;
              break;
            }
          if (n > 0)
            return n;
        }
      }
      return 0;
    }
    else
    {
      int r = cell - 1;
      if (surviveOperatorSelector.equals("OR >="))
      {
        for (int c = 0; c < columns; c++)
          if ((survive[r][c] > 0) && (counts[c] >= survive[r][c]))
            return cell;
      }
      else if (surviveOperatorSelector.equals("OR ="))
      {
        for (int c = 0; c < columns; c++)
          if ((survive[r][c] > 0) && (counts[c] == survive[r][c]))
            return cell;
      }
      else if (surviveOperatorSelector.equals("AND >="))
      {
        int n = 0;
        for (int c = 0; c < columns; c++)
          if (survive[r][c] > 0)
          {
            n = cell;
            break;
          }
        if (n == 0)
          return 0;
        for (int c = 0; c < columns; c++)
          if ((survive[r][c] > 0) && (counts[c] < survive[r][c]))
            return 0;
        return n;
      }
      else if (surviveOperatorSelector.equals("AND ="))
      {
        int n = 0;
        for (int c = 0; c < columns; c++)
          if (survive[r][c] > 0)
          {
            n = cell;
            break;
          }
        if (n == 0)
          return 0;
        for (int c = 0; c < columns; c++)
          if ((survive[r][c] > 0) && (counts[c] != survive[r][c]))
            return 0;
        return n;
      }
      return 0;
    }
  }

  public int getNoStates()
  {
    return noStates;
  }

  public int getBirth(int row, int column)
  {
    return birth[row][column];
  }

  public int getSurvive(int row, int column)
  {
    return survive[row][column];
  }

  public int getBirthRandomLimit()
  {
    return birthRandomLimit;
  }

  public int getSurviveRandomLimit()
  {
    return surviveRandomLimit;
  }

  public String getBirthOperatorSelector()
  {
    return birthOperatorSelector;
  }

  public String getSurviveOperatorSelector()
  {
    return surviveOperatorSelector;
  }

  public void setBirth(int row, int column, int value)
  {
    birth[row][column] = value;
  }

  public void setSurvive(int row, int column, int value)
  {
    survive[row][column] = value;
  }

  public void setBirthRandomLimit(int value)
  {
    birthRandomLimit = value;
  }

  public void setSurviveRandomLimit(int value)
  {
    surviveRandomLimit = value;
  }

  public void setBirthOperatorSelector(String value)
  {
    birthOperatorSelector = value;
  }

  public void setSurviveOperatorSelector(String value)
  {
    surviveOperatorSelector = value;
  }

  public void setBirthEnabled(boolean value)
  {
    birthEnabled = value;
  }

  public boolean getBirthEnabled()
  {
    return birthEnabled;
  }

  public void setSurviveEnabled(boolean value)
  {
    surviveEnabled = value;
  }

  public boolean getSurviveEnabled()
  {
    return surviveEnabled;
  }

  public void setBirthPercentChange(int value)
  {
    birthPercentChange = value;
  }

  public int getBirthPercentChange()
  {
    return birthPercentChange;
  }

  public void setSurvivePercentChange(int value)
  {
    survivePercentChange = value;
  }

  public int getSurvivePercentChange()
  {
    return survivePercentChange;
  }

  public void setBirthClear(boolean value)
  {
    birthClear = value;
  }

  public boolean getBirthClear()
  {
    return birthClear;
  }

  public void setSurviveClear(boolean value)
  {
    surviveClear = value;
  }

  public boolean getSurviveClear()
  {
    return surviveClear;
  }
}
