import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

class RuleTableRule extends Rule
{
  private static final int noStateWeights = 16;
  private static final int neighborRows = 3;
  private static final int neighborColumns = 5;

  private int noStates;
  private int noCounts;
  private int percentChange = 30;
  private String newRuleOption = "Random";
  private boolean clearNewRule = true;
  private int stateWeights[];
  private int neighborWeights[][];
  private int ruleTable[][];
  private int useTable[][];

  public RuleTableRule(JFrame mainFrame, TriangularCell triangularCell)
  {
    super(mainFrame, triangularCell);
    setDefault();
  }

  private void makeArrays()
  {
    stateWeights = new int[noStateWeights];
    neighborWeights = new int[neighborRows][neighborColumns];
    ruleTable = new int[noStates][noCounts];
    useTable = new int[noStates][noCounts];
  }

  private void setWeights()
  {
    for (int r = 0; r < noStateWeights; r++)
      stateWeights[r] = 0;
    stateWeights[1] = 1;
    for (int r = 0; r < neighborRows; r++)
      for (int c = 0; c < neighborColumns; c++)
        neighborWeights[r][c] = 0;
    neighborWeights[0][2] = 1;
    neighborWeights[1][1] = 1;
    neighborWeights[1][3] = 1;
  }

  private void wireWorld()
  {
    noStates = 4;
    noCounts = 9;
    makeArrays();
    setWeights();
    for (int r = 0; r < noStates; r++)
      for (int c = 0; c < noCounts; c++)
        if (r == 1)
          ruleTable[r][c] = 2;
        else if ((r == 2) || (r == 3))
          ruleTable[r][c] = 3;
        else
          ruleTable[r][c] = 0;
    ruleTable[3][1] = 1;
    ruleTable[3][2] = 1;
  }

  public void setDefault()
  {
    wireWorld();
  }

  public void changeRule()
  {
    RuleTableDialog ruleTableDialog = new RuleTableDialog(mainFrame, this, triangularCell);
    ruleTableDialog.setData(this);
    centerDialog(ruleTableDialog);
    ruleTableDialog.setVisible(true);
    if (ruleTableDialog.hasChanged())
		{
      stateWeights = new int[noStateWeights];
      ruleTableDialog.getData(this);
		}
  }

  private void centerDialog(JDialog dialog)
  {
    Point p = mainFrame.getLocation();
    int x = (int)p.getX() + (mainFrame.getWidth() - dialog.getWidth()) / 2;
    int y = (int)p.getY() + (mainFrame.getHeight() - dialog.getHeight()) / 2;
    dialog.setLocation(x, y);
  }

  private void clearUseTable()
  {
    for (int r = 0; r < noStates; r++)
      for (int c = 0; c < noCounts; c++)
        useTable[r][c] = 0;
  }

  public void setNewRule()
  {
    random(ruleTable);
  }

  public void random(int ruleTable[][])
  {
    clearUseTable();
    if (clearNewRule)
    {
      for (int r = 0; r < noStates; r++)
        for (int c = 0; c < noCounts; c++)
          ruleTable[r][c] = 0;
    }
    Random generator = new Random();
    if (newRuleOption.equals("Random"))
    {
      for (int i = 1; i < noStates * noCounts * percentChange / 100; i++)
      {
        int r = generator.nextInt(noStates);
        int c = generator.nextInt(noCounts);
        ruleTable[r][c] = generator.nextInt(noStates);
      }
    }
    else if (newRuleOption.equals("Generations") || newRuleOption.equals("Modified Generations"))
    {
      for (int r = 1; r < (noStates - 1); r++)
        for (int c = 0; c < noCounts; c++)
          ruleTable[r][c] = r + 1;
      if (newRuleOption.equals("Generations"))
        for (int i = 1; i < 2 * noCounts * percentChange / 100; i++)
        {
          int r = generator.nextInt(2);
          int c = generator.nextInt(noCounts);
          ruleTable[r][c] = 1;
        }
      else
        for (int i = 1; i < noStates * noCounts * percentChange / 100; i++)
        {
          int r = generator.nextInt(noStates);
          int c = generator.nextInt(noCounts);
          ruleTable[r][c] = generator.nextInt(noStates);
        }
    }
  }

  public void setOptions()
  {
    int currentNoStates = noStates;
    int currentNoCounts = noCounts;
    RuleTableOptionsDialog ruleTableOptionsDialog = new RuleTableOptionsDialog(mainFrame);
    ruleTableOptionsDialog.setData(this);
    centerDialog(ruleTableOptionsDialog);
    ruleTableOptionsDialog.setVisible(true);
    if (ruleTableOptionsDialog.hasChanged())
    {
      ruleTableOptionsDialog.getData(this);
      if ((noStates != currentNoStates) || (noCounts != currentNoCounts))
      {
				if (noStates != currentNoStates)
					for (int i = noStates; i < noStateWeights; i++)
					  stateWeights[i] = 0;
        ruleTable = new int[noStates][noCounts];
        useTable = new int[noStates][noCounts];
      }
    }
  }

  public void writeRule(PrintWriter pw)
  {
    pw.println("#Rule = Rule Table");
    pw.println("#States = " + noStates);
    pw.println("#Counts = " + noCounts);
    String s = "#SW ";
    for (int i = 0; i < noStates; i++)
      s = s + stateWeights[i] + ",";
    pw.println(s.substring(0, s.lastIndexOf(",")));
    for (int i = 0; i < neighborRows; i++)
    {
      s = "#NW ";
      for (int j = 0; j < neighborColumns; j++)
        s = s + neighborWeights[i][j] + ",";
      pw.println(s.substring(0, s.lastIndexOf(",")));
    }
    for (int i = 0; i < noStates; i++)
    {
      s = "#RT ";
      for (int j = 0; j < noCounts; j++)
        s = s + ruleTable[i][j] + ",";
      pw.println(s.substring(0, s.lastIndexOf(",")));
    }
  }

  public void readRule(BufferedReader br)
  {
    try
    {
      noStates = new Integer(br.readLine().substring(10).trim());
      noCounts = new Integer(br.readLine().substring(10).trim());
      ruleTable = new int[noStates][noCounts];
      useTable = new int[noStates][noCounts];
			for (int i = 0; i < noStateWeights; i++)
				stateWeights[i] = 0;
      String values[] = br.readLine().substring(3).split(",");
      for (int i = 0; i < noStates; i++)
        stateWeights[i] = new Integer(values[i].trim());
      for (int i = 0; i < neighborRows; i++)
      {
        values = br.readLine().substring(3).split(",");
        for (int j = 0; j < neighborColumns; j++)
          neighborWeights[i][j] = new Integer(values[j].trim());
      }
      for (int i = 0; i < noStates; i++)
      {
        values = br.readLine().substring(3).split(",");
        for (int j = 0; j < noCounts; j++)
          ruleTable[i][j] = new Integer(values[j].trim());
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  public int step(int c, int r)
  {
    int cell = triangularCell.getCell(c, r);
    int count;
    if (!((c & 1) == 0) ^ ((r & 1) == 0))
    {
      count =
      stateWeights[triangularCell.getCell(c - 2, r + 1)] * neighborWeights[0][4] +
      stateWeights[triangularCell.getCell(c + 2, r + 1)] * neighborWeights[0][0] +
      stateWeights[triangularCell.getCell(c    , r - 1)] * neighborWeights[2][2] +
      stateWeights[triangularCell.getCell(c    , r + 1)] * neighborWeights[0][2] +
      stateWeights[triangularCell.getCell(c - 1, r - 1)] * neighborWeights[2][3] +
      stateWeights[triangularCell.getCell(c - 1, r    )] * neighborWeights[1][3] +
      stateWeights[triangularCell.getCell(c - 1, r + 1)] * neighborWeights[0][3] +
      stateWeights[triangularCell.getCell(c + 1, r    )] * neighborWeights[1][1] +
      stateWeights[triangularCell.getCell(c + 1, r + 1)] * neighborWeights[0][1] +
      stateWeights[triangularCell.getCell(c + 2, r    )] * neighborWeights[1][0] +
      stateWeights[triangularCell.getCell(c - 2, r    )] * neighborWeights[1][4] +
      stateWeights[triangularCell.getCell(c + 1, r - 1)] * neighborWeights[2][1] +
      stateWeights[cell] * neighborWeights[1][2];
    }
    else
    {
      count =
      stateWeights[triangularCell.getCell(c - 2, r - 1)] * neighborWeights[0][0] +
      stateWeights[triangularCell.getCell(c + 2, r - 1)] * neighborWeights[0][4] +
      stateWeights[triangularCell.getCell(c    , r - 1)] * neighborWeights[0][2] +
      stateWeights[triangularCell.getCell(c    , r + 1)] * neighborWeights[2][2] +
      stateWeights[triangularCell.getCell(c - 1, r - 1)] * neighborWeights[0][1] +
      stateWeights[triangularCell.getCell(c - 1, r    )] * neighborWeights[1][1] +
      stateWeights[triangularCell.getCell(c - 1, r + 1)] * neighborWeights[2][1] +
      stateWeights[triangularCell.getCell(c + 1, r    )] * neighborWeights[1][3] +
      stateWeights[triangularCell.getCell(c + 1, r + 1)] * neighborWeights[2][3] +
      stateWeights[triangularCell.getCell(c + 2, r    )] * neighborWeights[1][4] +
      stateWeights[triangularCell.getCell(c - 2, r    )] * neighborWeights[1][0] +
      stateWeights[triangularCell.getCell(c + 1, r - 1)] * neighborWeights[0][3] +
      stateWeights[cell] * neighborWeights[1][2];
    }
    if ((count >= 0) && (count < noCounts))
    {
      useTable[cell][count] = 1;
      return ruleTable[cell][count];
    }
    return 0;
  }

  public int getNoStates()
  {
    return noStates;
  }

  public int getNoCounts()
  {
    return noCounts;
  }

  public int getPercentChange()
  {
    return percentChange;
  }

  public String getNewRuleOption()
  {
    return newRuleOption;
  }

  public boolean getClearNewRule()
  {
    return clearNewRule;
  }

  public int getStateWeight(int state)
  {
    return stateWeights[state];
  }

  public int getNeighborWeight(int row, int column)
  {
    return neighborWeights[row][column];
  }

  public int getRuleTableValue(int row, int column)
  {
    return ruleTable[row][column];
  }

  public int getUseTableValue(int row, int column)
  {
    return useTable[row][column];
  }

  public void setNoStates(int noStates)
  {
    this.noStates = noStates;
  }

  public void setNoCounts(int noCounts)
  {
    this.noCounts = noCounts;
  }

  public void setPercentChange(int percentChange)
  {
    this.percentChange = percentChange;
  }

  public void setNewRuleOption(String newRuleOption)
  {
    this.newRuleOption = newRuleOption;
  }

  public void setClearNewRule(boolean clearNewRule)
  {
    this.clearNewRule = clearNewRule;
  }

  public void setStateWeight(int state, int value)
  {
    stateWeights[state] = value;
  }

  public void setNeighborWeight(int row, int column, int value)
  {
    neighborWeights[row][column] = value;
  }

  public void setRuleTableValue(int row, int column, int value)
  {
    ruleTable[row][column] = value;
  }

  public void setUseTableValue(int row, int column, int value)
  {
    useTable[row][column] = value;
  }
}
