import java.awt.*;
import java.io.*;

public class Selector
{
  public static int stateMask = 31;
  public static int selected = 32;

  public enum State {None, TopLeftMarked, GroupSelected};

  private TriangularCell triangularCell;
  private TriGrid triGrid;
  private int cells[][];
  private int worldColumns;
  private int worldRows;
  private int selection[][];
  private State state;
  private int topLeftColumn;
  private int topLeftRow;
  private int columns;
  private int rows;
  private boolean pointingUp;
  private boolean cut;

  public Selector(TriangularCell triangularCell, TriGrid triGrid, int cells[][], int noColumns, int noRows)
  {
    this.triangularCell = triangularCell;
    this.triGrid = triGrid;
    this.cells = cells;
    worldColumns = noColumns;
    worldRows = noRows;
    state = State.None;
    cut = true;
  }

  public void setCells(int[][] cells)
  {
    this.cells = cells;
    if (state == State.GroupSelected)
      markSelectedCells();
  }

  public int getColumns()
  {
    return columns;
  }

  public int getRows()
  {
    return rows;
  }

  public State getState()
  {
    return state;
  }

  public void setCut(boolean cut)
  {
    this.cut = cut;
  }

  public boolean getCut()
  {
    return cut;
  }

  public int getCell(int i, int j)
  {
    return selection[(worldColumns + i - topLeftColumn) % worldColumns][(worldRows + j - topLeftRow) % worldRows];
  }

  public void setTopLeft(int c, int r)
  {
    Point p = triGrid.selectCell(c, r);
    if (p == null)
      return;
    state = State.TopLeftMarked;
    topLeftColumn = p.x;
    topLeftRow = p.y;
    pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
    makeNewSelection(p.x, p.y);
    selectCells(false);
    triGrid.setDrawSelector(true);
    triGrid.repaint();
  }

  public void makeNewSelection(int column, int row)
  {
    int t;
    int c = topLeftColumn;
    int r = topLeftRow;
    if (c > column)
    {
      t = c;
      c = column;
      column = t;
    }
    if (r > row)
    {
      t = r;
      r = row;
      row = t;
    }
    topLeftColumn = c;
    topLeftRow = r;
    pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
    columns = column - topLeftColumn + 1;
    rows = row - topLeftRow + 1;
    selection = new int[columns][rows];
  }

  private void selectCells(boolean cut)
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
    {
      int c = triGrid.offsetColumn(i);
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
      {
        int r = triGrid.offsetRow(j);
        selection[i - topLeftColumn][j - topLeftRow] = cells[c][r] & stateMask;
        if (cut)
          cells[c][r] = selected;
        else
          cells[c][r] |= selected;
      }
    }
  }

  public void save(int c, int r)
  {
    Point p = triGrid.selectCell(c, r);
    if (p == null)
      return;
    triGrid.setDrawSelector(false);
    state = State.GroupSelected;
    makeNewSelection(p.x, p.y);
    selectCells(cut);
    triGrid.setDrawSelector(true);
    triGrid.repaint();
  }

  private void clearSelectedBits()
  {
    for (int i = 0; i < worldColumns; i++)
      for (int j = 0; j < worldRows; j++)
        cells[i][j] &= stateMask;
  }

  public void restoreCells()
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
    {
      int c = triGrid.offsetColumn(i);
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
      {
        int r = triGrid.offsetRow(j);
        int v = selection[i - topLeftColumn][j - topLeftRow];
        if (v != 0)
          cells[c][r] = v;
      }
    }
  }

  public void remove()
  {
    triGrid.setDrawSelector(false);
    state = State.None;
    clearSelectedBits();
    selection = null;
    triGrid.repaint();
  }

  public void clearOutside()
  {
    for (int c = 0; c < worldColumns; c++)
      for (int r = 0; r < worldRows; r++)
        cells[c][r] &= selected;
    triGrid.repaint();
  }

  private void markSelectedCells()
  {
    for (int i = topLeftColumn; i < topLeftColumn + columns; i++)
      for (int j = topLeftRow; j < topLeftRow + rows; j++)
        cells[triGrid.offsetColumn(i)][triGrid.offsetRow(j)] |= selected;
  }

  public boolean move(int c, int r)
  {
    Point p = triGrid.selectCell(c, r);
    if (p == null)
      return false;
    if (triGrid.isPointingUp(p.x, p.y) != pointingUp)
      return false;
    triGrid.setDrawSelector(false);
    topLeftColumn = p.x;
    topLeftRow = p.y;
    clearSelectedBits();
    markSelectedCells();
    triGrid.setDrawSelector(true);
    triGrid.repaint();
    return true;
  }

  private void trimSelection()
  {
    int leftColumn = columns;
    int topRow = rows;
    int rightColumn = -1;
    int bottomRow = -1;
    for (int c = 0; c < columns; c++)
      for (int r = 0; r < rows; r++)
        if (selection[c][r] != 0)
        {
          if (c < leftColumn)
            leftColumn = c;
          if (c > rightColumn)
            rightColumn = c;
          if (r < topRow)
            topRow = r;
          if (r > bottomRow)
            bottomRow = r;
        }
    int newColumns = rightColumn - leftColumn + 1;
    int newRows = bottomRow - topRow + 1;
    int newSelection[][] = new int[newColumns][newRows];
    for (int c = 0; c < newColumns; c++)
      for (int r = 0; r < newRows; r++)
        newSelection[c][r] = selection[c + leftColumn][r + topRow];
    incTopLeftColumn(leftColumn);
    incTopLeftRow(topRow);
    pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
    selection = newSelection;
    columns = newColumns;
    rows = newRows;
  }

  public void flipX()
  {
    if (state != State.GroupSelected)
      return;
    triGrid.setDrawSelector(false);
    int t[][];
    if ((rows % 2) == 0)
    {
      t = new int[columns][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[c][rows - r - 1] = selection[c][r];
    }
    else
    {
      t = new int[columns][rows + 1];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[c][rows - r] = selection[c][r];
      rows++;
    }
    selection = t;
    trimSelection();
    center();
  }

  public void flipY()
  {
    if (state != State.GroupSelected)
      return;
    int t[][];
    triGrid.setDrawSelector(false);
    if ((columns % 2) == 1)
    {
      t = new int[columns][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[columns - c - 1][r] = selection[c][r];
    }
    else
    {
      t = new int[columns + 1][rows];
      for (int c = 0; c < columns; c++)
        for (int r = 0; r < rows; r++)
          t[columns - c][r] = selection[c][r];
      columns++;
    }
    selection = t;
    trimSelection();
    center();
  }

  public void rotateClockwise()
  {
    if (state != State.GroupSelected)
      return;
    int newSelection[][], newColumns, newRows, sc;
    if (pointingUp)
    {
      newColumns = columns / 2 + (rows / 2) * 3 + (rows % 2) * 2 - 1;
      newRows = rows / 2 + columns / 2 + (columns % 2) * (rows % 2);
      newSelection = new int[newColumns][newRows];
      int ic = (rows / 2) * 3 + (rows % 2) * 2 - 2;
      sc = ic;
      int ir = 0;
      for (int r = 0; r < rows; r++)
      {
        int nc = ic;
        int nr = ir;
        for (int c = 0; c < columns; c++)
        {
          newSelection[nc][nr] = selection[c][r];
          if (!((c % 2) == 0) ^ ((r % 2) == 0))
            nc++;
          else
            nr++;
        }
        if ((r % 2) == 0)
          ic--;
        else
        {
          ic -= 2;
          ir++;
        }
      }
    }
    else
    {
      newColumns = (columns - 1) / 2 + (rows / 2) * 3 + rows % 2;
      newRows = rows / 2 + columns / 2;
      if ((columns % 2) == 0)
        newRows += (rows % 2);
      else
        newRows++;
      newSelection = new int[newColumns][newRows];
      int ic = (rows / 2) * 3 + rows % 2 - 1;
      sc = ic;
      int ir = 0;
      for (int r = 0; r < rows; r++)
      {
        int nc = ic;
        int nr = ir;
        for (int c = 0; c < columns; c++)
        {
          newSelection[nc][nr] = selection[c][r];
          if (((c % 2) == 0) ^ ((r % 2) == 0))
            nc++;
          else
            nr++;
        }
        if ((r % 2) == 1)
          ic--;
        else
        {
          ic -= 2;
          ir++;
        }
      }
    }
    triGrid.setDrawSelector(false);
    if ((sc % 2) == 0)
      incTopLeftColumn(1);
    pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
    selection = newSelection;
    columns = newColumns;
    rows = newRows;
    trimSelection();
    center();
  }

  public void rotateAntiClockwise()
  {
    if (state != State.GroupSelected)
      return;
    int newSelection[][], newColumns, newRows, sr;
    if (pointingUp)
    {
      newColumns = columns / 2 + columns % 2 - 1;
      if ((columns % 2) == 0)
        newColumns += (rows / 2) * 3 + rows % 2;
      else
        newColumns += ((rows - 1) / 2) * 3 + (rows - 1) % 2 + 1;
      newRows = columns / 2 + rows / 2 + rows % 2;
      newSelection = new int[newColumns][newRows];
      int ic = 0;
      int ir = (columns - 1) / 2 + (columns - 1) % 2;
      sr = ir;
      for (int r = 0; r < rows; r++)
      {
        int nc = ic;
        int nr = ir;
        for (int c = 0; c < columns; c++)
        {
          newSelection[nc][nr] = selection[c][r];
          if (!((c % 2) == 0) ^ ((r % 2) == 0))
            nr--;
          else
            nc++;
        }
        if ((r % 2) == 0)
          ic++;
        else
        {
          ic += 2;
          ir++;
        }
      }
    }
    else
    {
      newColumns = columns / 2;
      if ((columns % 2) == 0)
        newColumns += ((rows - 1) / 2) * 3 + (rows - 1) % 2 + 1;
      else
        newColumns += (rows / 2) * 3 + rows % 2;
      newRows = columns / 2 + columns % 2 + rows / 2;
      newSelection = new int[newColumns][newRows];
      int ic = 0;
      int ir = (columns - 1) / 2;
      sr = ir;
      for (int r = 0; r < rows; r++)
      {
        int nc = ic;
        int nr = ir;
        for (int c = 0; c < columns; c++)
        {
          newSelection[nc][nr] = selection[c][r];
          if (!((c % 2) == 0) ^ ((r % 2) == 0))
            nc++;
          else
            nr--;
        }
        if ((r % 2) == 0)
        {
          ic += 2;
          ir++;
        }
        else
          ic++;
      }
    }
    triGrid.setDrawSelector(false);
    if ((sr % 2) == 0)
      incTopLeftColumn(1);
    pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
    selection = newSelection;
    columns = newColumns;
    rows = newRows;
    trimSelection();
    center();
  }

  public void selectAll()
  {
    if (state != State.None)
    {
      restoreCells();
      remove();
    }
    int c1 = worldColumns;
    int r1 = worldRows;
    int cn = 0;
    int rn = 0;
    boolean clear = true;
    for (int c = 0; c < worldColumns; c++)
      for (int r = 0; r < worldRows; r++)
        if (cells[c][r] > 0)
        {
          clear = false;
          if (c < c1)
            c1 = c;
          if (c > cn)
            cn = c;
          if (r < r1)
            r1 = r;
          if (r > rn)
            rn = r;
        }
    if (! clear)
    {
      topLeftColumn = triGrid.worldColumnToGridColumn(c1);
      topLeftRow = triGrid.worldRowToGridRow(r1);
      pointingUp = triGrid.isPointingUp(topLeftColumn, topLeftRow);
      columns = triGrid.worldColumnToGridColumn(cn) - topLeftColumn + 1;
      if (columns <= 0)
        columns += worldColumns;
      rows = triGrid.worldRowToGridRow(rn) - topLeftRow + 1;
      if (rows <= 0)
        rows += worldRows;
      selection = new int[columns][rows];
      selectCells(cut);
      state = State.GroupSelected;
      triGrid.setDrawSelector(true);
      triGrid.repaint();
    }
  }

  public void center()
  {
    if ((state != State.GroupSelected) || (columns > triGrid.getGridColumns()) || (rows > triGrid.getGridRows()))
      return;
    triGrid.setDrawSelector(false);
    topLeftColumn = (triGrid.getGridColumns() - columns) / 2;
    topLeftRow = (triGrid.getGridRows() - rows) / 2;
    if (triGrid.isPointingUp(topLeftColumn, topLeftRow) != pointingUp)
      decTopLeftRow(1);
    clearSelectedBits();
    markSelectedCells();
    triGrid.setDrawSelector(true);
    triGrid.repaint();
  }

  public void incTopLeftColumn(int inc)
  {
    topLeftColumn = (topLeftColumn + inc) % worldColumns;
  }

  public void decTopLeftColumn(int inc)
  {
    topLeftColumn = (worldColumns + topLeftColumn - inc) % worldColumns;
  }

  public void incTopLeftRow(int inc)
  {
    topLeftRow = (topLeftRow + inc) % worldRows;
  }

  public void decTopLeftRow(int inc)
  {
    topLeftRow = (worldRows + topLeftRow - inc) % worldRows;
  }

  public void readNewSelection(BufferedReader br)
  {
    if (state != State.None)
    {
      restoreCells();
      remove();
    }
    try
    {
      rows = new Integer(br.readLine().substring(8).trim());
      columns = new Integer(br.readLine().substring(11).trim());
      if (br.readLine().substring(15).equals("Up"))
      {
        pointingUp = true;
        topLeftRow = 0;
      }
      else
      {
        pointingUp = false;
        topLeftRow = 1;
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    topLeftColumn = 0;
    selection = new int[columns][rows];
    readSelection(br);
    state = State.GroupSelected;
    center();
  }

  private void readSelection(BufferedReader br)
  {
    int r, c;
    int n;
    char ch;
    String s = "";
    try
    {
      String line = br.readLine();
      while (line != null)
        if (line.startsWith("#L "))
        {
          s = s + line.substring(3);
          line = br.readLine();
        }
        else
          return;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    r = 0;
    c = 0;
    while (s.length() > 0)
    {
      n = 0;
      ch = s.charAt(0);
      while ((ch >= '0') && (ch <= '9'))
      {
        n = n * 10 + (ch - '0');
        s = s.substring(1);
        ch = s.charAt(0);
      }
      if (n == 0)
        n = 1;
      if (ch == '$')
      {
        r += n;
        c = 0;
      }
      else
        if (ch == '.')
          c += n;
        else
        {
          for (int i = 0;  i < n; i++)
          {
            selection[c][r] = (byte)(ch - '@');
            c++;
          }
        }
      s = s.substring(1);
    }
  }

  public void writeHeader(PrintWriter pw)
  {
    pw.println("#Rows = " + rows);
    pw.println("#Columns = " + columns);
    if (pointingUp)
      pw.println("#Orientation = Up");
    else
      pw.println("#Orientation = Down");
  }

  public void writeSelection(PrintWriter pw)
  {
    int r, c;
    int ce = columns - 1;
    char state;
    char lastState = '0';
    int n = 0;
    boolean blankLine;
    String s = "";
    for (r = 0; r < rows; r++)
    {
      blankLine = true;
      for (c = 0; c < columns; c++)
      {
        if (selection[c][r] > 0)
        {
          blankLine = false;
          break;
        }
      }
      if (blankLine)
      {
        if (lastState == '$')
          n++;
        else
        {
          lastState = '$';
          n = 1;
        }
        continue;
      }
      else
        for (c = columns - 1; c >= 0; c--)
        {
          if (selection[c][r] > 0)
          {
            ce = c;
            break;
          }
        }
      for (c = 0; c <= ce;  c++)
      {
        state = (char)(selection[c][r] + 64);
        if (state == '@')
          state = '.';
        if (n == 0)
        {
          lastState = state;
          n = 1;
        }
        else if (state == lastState)
          n++;
        else
        {
          if (n > 1)
            s = s + n;
          s = s + lastState;
          if (s.length() >= 67)
          {
            pw.println("#L " + s);
            s = "";
          }
          lastState = state;
          n = 1;
        }
      }
      if (lastState == '$')
        n++;
      else
      {
        if (n > 1)
          s = s + n;
        s = s + lastState;
        if (s.length() >= 67)
        {
          pw.println("#L " + s);
          s = "";
        }
        lastState = '$';
        n = 1;
      }
    }
    if (s != "")
      pw.println("#L " + s);
  }
  
  public void random()
  {
    if (state != State.GroupSelected)
      return;
    triangularCell.random(columns, rows, selection);
  }
}