import java.awt.*;
import javax.swing.*;
import java.awt.image.BufferedImage;

class TriGrid extends JPanel
{
  private int cells[][];
  private int worldColumns;
  private int worldRows;
  private boolean gridValid;
  private int gridColumns;
  private int gridRows;
  private int xOffset;
  private int yOffset;
  private int cellWidth;
  private int cellHeight;
  private int gridWidth;
  private int gridHeight;
  private int border;
  private int columnOffset;
  private int rowOffset;
  private int zoomLevel;
  private BufferedImage image;
  private TriangularCell triangularCell;
  private Selector selector;
  private StatusBar statusBar;
  private ColorSelector colorSelector;
  private int currentState;
  private boolean drawSelector;

  public TriGrid(int cells[][], int noColumns, int noRows)
  {
    setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
    this.cells = cells;
    worldColumns = noColumns;
    worldRows = noRows;
    currentState = 1;
    gridValid = false;
    columnOffset = 0;
    rowOffset = 0;
    zoomLevel = 10;
    border = 10;
    drawSelector = false;
  }

  public void setComponents(TriangularCell triangularCell, Selector selector, StatusBar statusBar, ColorSelector colorSelector)
  {
    this.triangularCell = triangularCell;
    this.selector = selector;    
    this.statusBar = statusBar;
    this.colorSelector = colorSelector;
  }

  public void setCells(int[][] cells)
  {
    this.cells = cells;
  }

  public void makeGrid()
  {
    cellWidth = 2 * zoomLevel;
    cellHeight = 2 * zoomLevel;
    gridWidth = getWidth() - 2 * border;
    gridHeight = getHeight() - 2 * border;
    gridColumns = (gridWidth - cellWidth / 2) / cellWidth * 2;
    if (gridColumns > worldColumns)
      gridColumns = worldColumns;
    xOffset = (getWidth() - (gridColumns * cellWidth + cellWidth) / 2) / 2;
    gridRows = (gridHeight / (2 * cellHeight)) * 2;
    if (gridRows > worldRows)
      gridRows = worldRows;
    yOffset = (getHeight() - gridRows * cellHeight) / 2;
    gridValid = true;
    statusBar.setGridSize(gridColumns, gridRows);
  }

  public void setDrawSelector(boolean value)
  {
    drawSelector = value;
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    if ((! gridValid) || (image == null) || (image.getWidth() != getWidth()) || (image.getHeight() != getHeight()))
    {
      image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
      makeGrid();
    }
    statusBar.setGridOffset(columnOffset, rowOffset);
    drawCells();
    if (drawSelector)
      drawSelectedCells();
    g.drawImage(image, 0, 0, null);
    if (! triangularCell.isRunning())
    {
      int p = 0;
       for (int c = 0; c < worldColumns; c++)
         for (int r = 0; r < worldRows; r++)
          if ((cells[c][r] & selector.stateMask) > 0)
            p++;
      statusBar.setPopulation(p);
    }
  }

  private Polygon getPolygon(int x, int y, boolean pointingUp)
  {
    int x1 = x + cellWidth / 2;
    int y1 = y + cellHeight;
    int x2 = x + cellWidth;
    Polygon polygon = new Polygon();
    if (pointingUp)
    {
      polygon.addPoint(x, y1);
      polygon.addPoint(x1, y);
      polygon.addPoint(x2, y1);
    }
    else
    {
      polygon.addPoint(x, y);
      polygon.addPoint(x2, y);
      polygon.addPoint(x1, y1);
    }
    return polygon;
  }

  public int offsetColumn(int c)
  {
    return (c + columnOffset) % worldColumns;
  }

  public int offsetRow(int r)
  {
    return (r + rowOffset) % worldRows;
  }

  public int worldColumnToGridColumn(int worldColumn)
  {
    return (worldColumns + worldColumn - columnOffset) % worldColumns;
  }

  public int worldRowToGridRow(int worldRow)
  {
    return (worldRows + worldRow - rowOffset) % worldRows;
  }

  public boolean isPointingUp(int c, int r)
  {
    int gridColumn = worldColumnToGridColumn(c);
    int gridRow = worldRowToGridRow(r);
    return !((gridColumn & 1) == 0) ^ ((gridRow & 1) == 0);
  }

  private void drawCells()
  {
    Graphics g = image.getGraphics();
    g.clearRect(0, 0, image.getWidth(), image.getHeight());
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      for (int j = 0; j < gridRows; j++)
      {
        Polygon p = getPolygon(xOffset + i * cellWidth / 2, yOffset + j * cellHeight, pointingUp);
        g.setColor(colorSelector.getColor(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask));
        g.fillPolygon(p);
        if (zoomLevel > 3)
        { 
          g.setColor(colorSelector.getGridColor());
          g.drawPolygon(p);
        }
        pointingUp = !pointingUp;
      }
    pointingUp = !pointingUp;
    }
    g.dispose();
  }


  private void drawSelectedCells()
  {
    Graphics g = image.getGraphics();
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      for (int j = 0; j < gridRows; j++)
      {
        if ((cells[offsetColumn(i)][offsetRow(j)] & selector.selected) != 0)
        {
          Polygon p = getPolygon(xOffset + i * cellWidth / 2, yOffset + j * cellHeight, pointingUp);
          int c = selector.getCell(i, j);
          if (c != 0)
          {
            g.setColor(colorSelector.getColor(c));
            g.fillPolygon(p);
          }
          g.setColor(colorSelector.getSelectorColor());
          g.drawPolygon(p);
        }
        pointingUp = !pointingUp;
      }
      pointingUp = !pointingUp;
    }
    g.dispose();
  }

  public void changeCurrentState()
  {
    currentState = (currentState + 1) % triangularCell.getNoStates();
  }

  public void setPosition(int x, int y)
  {
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      for (int j = 0; j < gridRows; j++)
      {
        if (getPolygon(xOffset + i * cellWidth / 2, yOffset + j * cellHeight, pointingUp).contains(x, y))
        {
          statusBar.setPosition(offsetColumn(i), offsetRow(j), i, j);
          statusBar.setState(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask);
          statusBar.repaint();
          return;
        }
        pointingUp = !pointingUp;
      }
    pointingUp = !pointingUp;
    }
    statusBar.removePosition();
    statusBar.removeState();
    statusBar.repaint();
  }

  public void setCellState(int x, int y)
  {
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      for (int j = 0; j < gridRows; j++)
      {
        if (getPolygon(xOffset + i * cellWidth / 2, yOffset + j * cellHeight, pointingUp).contains(x, y))
        {
          cells[offsetColumn(i)][offsetRow(j)] = currentState;
          statusBar.setState(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask);
          repaint();
          return;
        }
        pointingUp = !pointingUp;
      }
    pointingUp = !pointingUp;
    }
  }

  public Point selectCell(int x, int y)
  {
    boolean pointingUp = true;
    for (int i = 0; i < gridColumns; i++)
    {
      for (int j = 0; j < gridRows; j++)
      {
        if (getPolygon(xOffset + i * cellWidth / 2, yOffset + j * cellHeight, pointingUp).contains(x, y))
          return new Point(i, j);
        pointingUp = !pointingUp;
      }
      pointingUp = !pointingUp;
    }
    return null;
  }

  public void incColumnOffset(int inc)
  {
    columnOffset = (columnOffset + inc) % worldColumns;
    if (selector.getState() != Selector.State.None)
      selector.decTopLeftColumn(inc);
    repaint();
  }

  public void decColumnOffset(int inc)
  {
    columnOffset = (worldColumns + columnOffset - inc) % worldColumns;
    if (selector.getState() != Selector.State.None)
      selector.incTopLeftColumn(inc);
    repaint();
  }

  public void incRowOffset(int inc)
  {
    rowOffset = (rowOffset + inc) % worldRows;
    if (selector.getState() != Selector.State.None)
      selector.decTopLeftRow(inc);
    repaint();
  }

  public void decRowOffset(int inc)
  {
    rowOffset = (worldRows + rowOffset - inc) % worldRows;
    if (selector.getState() != Selector.State.None)
      selector.incTopLeftRow(inc);
    repaint();
  }

  public void resetOffset()
  {
    columnOffset = 0;
    rowOffset = 0;
  }

  public void zoomIn()
  {
    zoomLevel++;
    gridValid = false;
    repaint();
  }

  public void zoomOut()
  {
    if (zoomLevel > 1)
    {
      zoomLevel--;
      gridValid = false;
      repaint();
    }
  }

  public void setCurrentState(int state)
  {
    currentState = state;
  }

  public int getCurrentState()
  {
    return currentState;
  }

  public int getGridColumns()
  {
    return gridColumns;
  }

  public int getGridRows()
  {
    return gridRows;
  }
}